/*
 * Copyright (C) 2004 YOSHIFUJI Hideaki.
 * Copyright (C) 2004 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>

#define DEFAULT_PORT	"4000"
#define DEFAULT_BACKLOG	32
#define MAX_BACKLOG	1024
#define DEFAULT_LOOP	1000

char *host         = NULL;
char *port         = DEFAULT_PORT;
int backlog        = DEFAULT_BACKLOG;
unsigned long loop = DEFAULT_LOOP;

int main(int argc, char **argv) 
{
	const char *progname = argv[0];

	char *host         = NULL;
	char *port         = DEFAULT_PORT;
	int backlog        = DEFAULT_BACKLOG;
	unsigned long loop = DEFAULT_LOOP;

	struct addrinfo hints, *res0, *res;
	int c;
	int gai;

	char hostbuf[NI_MAXHOST], servbuf[NI_MAXSERV];

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;

	while((c = getopt(argc, argv, "46b:hl:p:s")) != EOF) {
		switch(c) {
		case '4':
			hints.ai_family = AF_INET;
			break;
		case '6':
			hints.ai_family = AF_INET6;
			break;
		case 'b':
		    do {
			unsigned long ul;
			ul = strtoul(optarg);
			if (ul > MAX_BACKLOG)
				ul = MAX_BACKLOG;
			backlog = (int)ul;
		    } while(0);
		case 'l':
			loop = strtoul(optarg);
			break;
		case 'p':
			port = optarg;
			break;
		case 's':
			hints.ai_flags |= AI_PASSIVE;
			break;
		case 'h':
			fprintf(stderr,
				"Usage: %s [-4|-6] [-p port] [-l] [host]\n",
				progname);
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;

	if (argc)
		host = argv[0];

	gai = getaddrinfo(host, port, &hints, &res0);
	if (gai) {
		fprintf(stderr,
			"getaddrinfo() failed for %s port %s: %s\n",
			host ? host : "<null>",
			port,
			gai_strerror(gai));
		exit(1);
	}

	if (hints.ai_flags & AI_PASSIVE) {
		/* server mode */
		int s = -1;
		unsigned long i;

		for (res = res0; res; res = res->ai_next) {
			s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
			if (s < 0) {
				fprintf(stderr,
					"socket: family %d, %s\n",
					res->ai_family, 
					strerror(errno));
				continue;
			}

			if (getnameinfo(res->ai_addr, res->ai_addrlen, 
					hostbuf, sizeof(hostbuf),
					servbuf, sizeof(servbuf),
					NI_NUMERICHOST|NI_NUMERICSERV)) {
				snprintf(hostbuf, sizeof(hostbuf), "??");
				snprintf(servbuf, sizeof(servbuf), "??");
			}

			if (bind(s, res->ai_addr, res->ai_addrlen) < 0) {
				fprintf(stderr,
					"bind: %s port %s, %s\n",
					hostbuf, servbuf, strerror(errno));
				close(s);
				s = -1;
				continue;
			}

			if (listen(s, backlog) < 0) {
				fprintf(stderr,
					"listen: %s port %s, %s\n",
					hostbuf, servbuf, strerror(errno));
				close(s);
				s = -1;
				continue;
			}
			break;
		}
		if (s < 0) {
			fprintf(stderr,
				"cannot listen on %s port %s\n",
				host ? host : "any",
				port);
			exit(1);
		}

		for (i = 0; !loop || i < loop; i++) {
			struct sockaddr_storage ss;
			socklen_t sslen = sizeof(ss);
			int newfd = accept(s, (struct sockaddr *)&ss, &sslen);
			if (newfd < 0) {
				printf("accept[%lu]: %s\n",
					i, strerror(errno));
			} else {
				if (getnameinfo((struct sockaddr *)&ss, sslen,
						hostbuf, sizeof(hostbuf),
						servbuf, sizeof(servbuf),
						NI_NUMERICHOST|NI_NUMERICSERV)) {
					snprintf(hostbuf, sizeof(hostbuf), "??");
					snprintf(servbuf, sizeof(servbuf), "??");
				}
				printf("accept[%lu]: %d, %s port %s\n",
					i, newfd, hostbuf, servbuf);
			}
		}
	} else {
		/* client mode */
		unsigned long i = 0;

		i = 0;
		res = res0;

		while ((!loop || i < loop) && res) {
			int s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
			if (s < 0) {
				printf("socket[%lu]: %s\n",
					i, strerror(errno));
				if (i == 0)
					res = res->ai_next;
				else
					i++;
				continue;
			}

			if (getnameinfo(res->ai_addr, res->ai_addrlen, 
					hostbuf, sizeof(hostbuf),
					servbuf, sizeof(servbuf),
					NI_NUMERICHOST|NI_NUMERICSERV)) {
				snprintf(hostbuf, sizeof(hostbuf), "??");
				snprintf(servbuf, sizeof(servbuf), "??");
			}

			if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
				printf("connect[%lu]: %s port %s, %s\n",
					i, hostbuf, servbuf, strerror(errno));
				close(s);
				if (i == 0)
					res = res->ai_next;
				else
					i++;
				continue;
			} else {
				printf("connect[%lu]: %s port %s, ok\n",
					i, hostbuf, servbuf, strerror(errno));
				close(s);
			}
			i++;
		}
	}
	printf("done.\n");
	exit(0);
}

